UNIT LRLRoutines;

INTERFACE

{$I LRL.HD1}

PROCEDURE VideoModeSet(Mode: BYTE);
PROCEDURE PaletteElementSet(Colour: BYTE; RedValue, GreenValue, BlueValue: BYTE);
PROCEDURE PaletteElementGet(Colour: BYTE; VAR RedValue, GreenValue, BlueValue: BYTE);
PROCEDURE Palette256Set(VAR Palette256);
PROCEDURE Palette256Get(VAR Palette256);
PROCEDURE Palette256Grayscale(VAR Palette256; StartElement, EndElement: BYTE);
PROCEDURE Palette256Darken(VAR Palette256; StartElement, EndElement, Decrement, MinValue: BYTE);
PROCEDURE Palette256Brighten(VAR Palette256; StartElement, EndElement, Increment, MaxValue: BYTE);
PROCEDURE Palette256Transform(VAR SourcePalette, DestinationPalette);
PROCEDURE RetraceWait;
FUNCTION VideoVGAInstalled: BOOLEAN;
FUNCTION DataByteGet(VAR Buffer; BufferOffset: WORD): BYTE;
PROCEDURE DataBytePut(VAR Buffer; BufferOffset: WORD; Value: BYTE);
FUNCTION DataWordGet(VAR Buffer; BufferOffset: WORD): WORD;
PROCEDURE DataWordPut(VAR Buffer; BufferOffset: WORD; Value: WORD);
PROCEDURE DataMove(VAR Source,Destination; Count: WORD; SourceOffset,DestinationOffset: WORD);
PROCEDURE DataAdd(VAR Buffer; Count: WORD; Amount: BYTE; BufferOffset: WORD);
PROCEDURE DataFill(VAR Buffer; Count: WORD; Value: BYTE; BufferOffset: WORD);
FUNCTION DataIdentical(VAR Array1,Array2; Count: WORD; Array1Offset, Array2Offset: WORD): BOOLEAN;
PROCEDURE DataExchange(VAR Array1,Array2; Count: WORD; Array1Offset,Array2Offset: WORD);
FUNCTION Readkey: WORD;
FUNCTION Keypressed: BOOLEAN;
FUNCTION SetInterrupt(Int: BYTE; NewAddress: POINTER): POINTER;
PROCEDURE CheckHardware;
PROCEDURE TimerFrequencySet(Freq: WORD);
PROCEDURE FadeClear;
PROCEDURE FadeTo(pal: POINTER);
PROCEDURE DecompressRepByte(VAR InArray,OutArray; InArraySize: WORD; VAR OutArraySize: WORD);
FUNCTION MSMouseInArea(x1,y1,x2,y2: INTEGER): BOOLEAN;
FUNCTION MSMouseDriverExist: BOOLEAN;
PROCEDURE MSMouseGetXY(VAR x,y: INTEGER);
FUNCTION MSMouseButtonStatusGet: WORD;
FUNCTION MSMouseButtonWasPressed(Button: WORD; VAR x,y: INTEGER): BOOLEAN;
FUNCTION MSMouseButtonWasReleased(Button: WORD; VAR x,y: INTEGER): BOOLEAN;
PROCEDURE MSMouseSetXY(x,y: INTEGER);
PROCEDURE KeyboardFlush;
PROCEDURE DisableInterrupts;
PROCEDURE EnableInterrupts;
PROCEDURE FHALT;
FUNCTION GetInterrupt(Int: BYTE): POINTER;

IMPLEMENTATION

{$I LRL.HDR}

VAR
	Old1B: POINTER;
	Old8: POINTER;

PROCEDURE NullInterrupt;
INTERRUPT;
ASSEMBLER;
ASM
END;

PROCEDURE VideoModeSet(Mode: BYTE);
ASSEMBLER;
ASM
	XOR AH,AH
	MOV AL,Mode
	INT 10H
END;

PROCEDURE PaletteElementSet(Colour: BYTE; RedValue, GreenValue, BlueValue: BYTE);
ASSEMBLER;
ASM
	MOV DX,3C8H
	MOV AL,Colour
	OUT DX,AL

	CLI
	MOV DX,3C9H
	MOV AL,RedValue
	OUT DX,AL
	MOV AL,GreenValue
	OUT DX,AL
	MOV AL,BlueValue
	OUT DX,AL
	STI
END;

PROCEDURE PaletteElementGet(Colour: BYTE; VAR RedValue, GreenValue, BlueValue: BYTE);
ASSEMBLER;
ASM
	MOV DX,3C7H
	MOV AL,Colour
	OUT DX,AL

	CLI
	MOV DX,3C9H
	IN AL,DX
	LES DI,RedValue
	STOSB
	IN AL,DX
	LES DI,GreenValue
	STOSB
	IN AL,DX
	LES DI,BlueValue
	STOSB
	STI
END;

PROCEDURE Palette256Set(VAR Palette256);
ASSEMBLER;
ASM
	PUSH DS

	MOV DX,3C8H
	MOV AL,0
	OUT DX,AL

	MOV CX,384

	LDS SI,Palette256

	MOV DX,3C9H

	CLI

@NextValue:
	LODSW
	OUT DX,AL
	XCHG AH,AL
	OUT DX,AL
	LOOP @NextValue

	STI

	POP DS
END;

PROCEDURE Palette256Get(VAR Palette256);
ASSEMBLER;
ASM
	PUSH DS

	MOV DX,3C7H
	MOV AL,0
	OUT DX,AL

	MOV CX,384

	LDS SI,Palette256

	CLI
	MOV DX,3C9H

@NextValue:
	IN AL,DX
	MOV AH,AL
	IN AL,DX
	XCHG AH,AL
	STOSW
	LOOP @NextValue
	STI

	POP DS
END;

PROCEDURE Palette256Grayscale(VAR Palette256; StartElement, EndElement: BYTE);
ASSEMBLER;
ASM
	PUSH DS

	LDS SI,Palette256
	LES DI,Palette256

	ADD SI,StartElement.WORD
	ADD DI,StartElement.WORD

	MOV CX,EndElement.WORD
	SUB CX,StartElement.WORD
	INC CX

	MOV DL,3

@Loop1:
	XOR AX,AX
	XOR BX,BX

	LODSB
	ADD BX,AX
	LODSB
	ADD BX,AX
	LODSB
	ADD BX,AX

	MOV AX,BX

	DIV DL

	STOSW
	STOSB

	LOOP @Loop1

	POP DS
END;

PROCEDURE Palette256Darken(VAR Palette256; StartElement, EndElement, Decrement, MinValue: BYTE);
ASSEMBLER;
ASM
	PUSH DS

	LDS SI,Palette256
	LES DI,Palette256

	MOV AX,StartElement.WORD
	MOV BL,3
	MUL BL

	MOV DX,AX
	ADD SI,AX
	ADD DI,AX

	MOV AX,EndElement.WORD
	MUL BL

	MOV CX,AX
	SUB CX,DX
	ADD CX,3

	XOR AX,AX

	MOV DX,Decrement.WORD
	MOV BX,MinValue.WORD

@Loop1:
	LODSB
	CMP AX,BX
	JBE @Next
	CMP AX,DX
	JB @SetMin
	SUB AX,DX
	JMP @Next

@SetMin:
	MOV AX,BX

@Next:
	STOSB
	LOOP @Loop1

	POP DS
END;

PROCEDURE Palette256Brighten(VAR Palette256; StartElement, EndElement, Increment, MaxValue: BYTE);
ASSEMBLER;
ASM
	PUSH DS

	LDS SI,Palette256
	LES DI,Palette256

	MOV AX,StartElement.WORD
	MOV BL,3
	MUL BL

	MOV DX,AX
	ADD SI,AX
	ADD DI,AX

	MOV AX,EndElement.WORD
	MUL BL

	MOV CX,AX
	SUB CX,DX
	ADD CX,3

	XOR AX,AX

	MOV DX,Increment.WORD
	MOV BX,MaxValue.WORD

@Loop1:
	XOR AX,AX
	LODSB
	CMP AX,BX
	JAE @Next
	ADD AX,DX
	CMP AX,BX
	JA @SetMax
	JMP @Next

@SetMax:
	MOV AX,BX

@Next:
	STOSB
	LOOP @Loop1

	POP DS
END;

PROCEDURE Palette256Transform(VAR SourcePalette, DestinationPalette);
ASSEMBLER;
ASM
	PUSH DS

	LDS SI,DestinationPalette
	PUSH DS
	PUSH SI
	LDS SI,SourcePalette
	LES DI,SourcePalette

	MOV CX,768

@Loop1:
	LODSB
	XCHG AH,AL

	MOV DX,SI
	MOV BX,DS
	POP SI
	POP DS
	LODSB
	PUSH DS
	PUSH SI
	MOV SI,DX
	MOV DS,BX

	CMP AH,AL
	JB @IncrementValue
	JE @Move
	DEC AH
	JMP @Move

@IncrementValue:
	INC AH

@Move:
	XCHG AH,AL
	STOSB
	LOOP @Loop1

	POP SI
	POP DS

	POP DS
END;

PROCEDURE RetraceWait;
ASSEMBLER;
ASM
	MOV DX,3DAH

@WaitForRetrace:
	IN AL,DX
	TEST AL,8
	JZ @WaitForRetrace
END;

FUNCTION VideoVGAInstalled: BOOLEAN;
ASSEMBLER;
ASM
	MOV AX,1A00H
	INT 10H
	CMP AL,1AH
	JNE @NotVGA
	CMP BL,7H
	JE @VGA
	CMP BL,8H
	JE @VGA
@NotVGA:
	MOV AL,0
	JMP @Done
@VGA:
	MOV AL,10000000B
@Done:
END;

FUNCTION DataByteGet(VAR Buffer; BufferOffset: WORD): BYTE;
ASSEMBLER;
ASM
	LES SI,Buffer
	ADD SI,BufferOffset
	MOV AL,[ES:SI]
END;

PROCEDURE DataBytePut(VAR Buffer; BufferOffset: WORD; Value: BYTE);
ASSEMBLER;
ASM
	LES DI,Buffer
	ADD DI,BufferOffset
	MOV AL,Value
	MOV [ES:DI],AL
END;

FUNCTION DataWordGet(VAR Buffer; BufferOffset: WORD): WORD;
ASSEMBLER;
ASM
	LES SI,Buffer
	ADD SI,BufferOffset
	MOV AX,[ES:SI]
END;

PROCEDURE DataWordPut(VAR Buffer; BufferOffset: WORD; Value: WORD);
ASSEMBLER;
ASM
	LES DI,Buffer
	ADD DI,BufferOffset
	MOV AX,Value
	MOV [ES:DI],AX
END;

PROCEDURE DataMove(VAR Source,Destination; Count: WORD; SourceOffset,DestinationOffset: WORD);
ASSEMBLER;
ASM
	MOV DX,DS

	LDS SI,Source
	ADD SI,SourceOffset
	LES DI,Destination
	ADD DI,DestinationOffset

	MOV CX,Count
	SHR CX,1
	JNC @MoveWords
	MOVSB

@MoveWords:
	REP MOVSW

	MOV DS,DX
END;

PROCEDURE DataFill(VAR Buffer; Count: WORD; Value: BYTE; BufferOffset: WORD);
ASSEMBLER;
ASM
	LES DI,Buffer
	ADD DI,BufferOffset

	MOV AL,Value
	MOV AH,AL

	MOV CX,Count
	SHR CX,1
	JNC @FillWords
	STOSB

@FillWords:
	REP STOSW
END;

FUNCTION DataIdentical(VAR Array1,Array2; Count: WORD; Array1Offset, Array2Offset: WORD): BOOLEAN;
ASSEMBLER;
ASM
	PUSH DS

	LDS SI,Array1
	ADD SI,Array1Offset
	LES DI,Array2
	ADD DI,Array2Offset

	XOR AX,AX

	MOV CX,Count
	SHR CX,1
	JNC @CompareWords
	CMPSB
	JNE @Done

@CompareWords:
	JCXZ @Done

	REPE CMPSW
	JNE @Done

	MOV AX,-1

@Done:
	POP DS
END;

PROCEDURE DataExchange(VAR Array1,Array2; Count: WORD; Array1Offset,Array2Offset: WORD);
ASSEMBLER;
ASM
	PUSH DS

	LDS SI,Array1
	ADD SI,Array1Offset
	LES DI,Array2
	ADD DI,Array2Offset

	MOV CX,Count
	SHR CX,1
	JNC @Label1
	LODSB
	MOV AH,[ES:DI]
	STOSB
	MOV [DS:SI-1],AH

@Label1:
	JCXZ @Done

@Loop1:
	LODSW
	MOV BX,[ES:DI]
	STOSW
	MOV [DS:SI-2],BX
	LOOP @Loop1

@Done:
	POP DS
END;

PROCEDURE DataAdd(VAR Buffer; Count: WORD; Amount: BYTE; BufferOffset: WORD);
ASSEMBLER;
ASM
	PUSH DS

	LES DI,Buffer
	ADD DI,BufferOffset
	MOV AX,ES
	MOV DS,AX
	MOV SI,DI

	MOV CX,Count
	MOV BL,Amount

	SHR CX,1
	JNC @DoWords

	LODSB
	ADD AL,BL
	STOSB

@DoWords:
	LODSW
	CMP AL,0
	JE @2
	ADD AL,BL

@2:
	CMP AH,0
	JE @3
	ADD AH,BL

@3:
	STOSW
	LOOP @DoWords

	POP DS
END;

FUNCTION Readkey: WORD;
ASSEMBLER;
ASM
	XOR AH,AH
	INT 16H
END;

FUNCTION Keypressed: BOOLEAN;
ASSEMBLER;
ASM
	XOR BX,BX
	MOV AH,01H
	INT 16H
	JZ @Done

	MOV BX,-1

@Done:
	MOV AX,BX
END;

FUNCTION SetInterrupt(Int: BYTE; NewAddress: POINTER): POINTER;
ASSEMBLER;
ASM
	MOV AL,Int
	MOV AH,35H
	INT 21H
	PUSH BX
	PUSH ES

	PUSH DS
	MOV AL,Int
	MOV AH,25H
	LDS DX,[NewAddress]
	INT 21H
	POP DS

	POP DX
	POP AX
END;

FUNCTION InfoCPUGet: BYTE; {from Intel (0 - 86, 2 - 286, 3 - 386, 4 - 486)}
ASSEMBLER;
ASM
@Check8086:
	PUSHF
	POP AX
	MOV CX,AX
	AND AX,0FFFH
	PUSH AX
	POPF
	PUSHF
	POP AX
	AND AX,0F000H
	MOV AX,0
	JNE @Check80286
	JNP @Done

@Check80286:
	OR CX,0F000H
	PUSH CX
	POPF
	PUSHF
	POP AX
	AND AX,0F000H
	MOV AX,2
	JNZ @Check80386
	JMP @Done

@Check80386:
	MOV BX,SP
	AND SP,NOT 3
	DB 66H; PUSHF
	DB 66H; POP AX
	DB 66H; MOV CX,AX
	DB 66H; DB 35H; DD 40000H
	DB 66H; PUSH AX
	DB 66H; POPF
	DB 66H; PUSHF
	DB 66H; POP AX
	DB 66H; XOR AX,CX
	MOV AX,3
	MOV SP,BX
	JZ @Done
	AND SP,NOT 3
	DB 66H; PUSH CX
	DB 66H; POPF
	MOV SP,BX

@Check80486:
	DB 66H; MOV AX,CX
	DB 66H; DB 35H; DD 200000H
	DB 66H; PUSH AX
	DB 66H; POPF
	DB 66H; PUSHF
	DB 66H; POP AX
	DB 66H; XOR AX,CX
	MOV AX,4
	JE @Done
	MOV AX,5

@Done:
END;

PROCEDURE CheckHardware;
BEGIN
	IF NOT VideoVGAInstalled THEN BEGIN
		WRITELN('Error: This program requires VGA!');
		WRITELN('訡: ⮩ ணࠬ 㦥  VGA!');
		HALT(2);
	END;
	IF InfoCPUGet<3 THEN BEGIN
		WRITELN('Error: This program requires i80386 processor or higher!');
		WRITELN('訡: ⮩ ணࠬ 㦥  i80386  !');
		HALT(3);
	END;
	IF MEMAVAIL<300000 THEN BEGIN
		WRITELN('Error: Not enough conventional memory!');
		WRITELN('訡:  墠⠥ ⨢ !');
		HALT(4);
	END;
END;

PROCEDURE TimerFrequencySet(Freq: WORD);
ASSEMBLER;
ASM
	MOV AL,00110110B
	OUT 43H,AL
	CMP Freq,18
	JA @Lab1
	XOR AX,AX
	JMP @Lab2

@Lab1:
	MOV BX,Freq
	MOV AX,0DC34H
	MOV DX,12H
	DIV BX

@Lab2:
	OUT 40H,AL
	XCHG AL,AH
	OUT 40H,AL
END;

PROCEDURE FadeClear;
VAR
	Pal1,Pal2: POINTER;
	i: INTEGER;
BEGIN
	GETMEM(Pal1,768);
	GETMEM(Pal2,768);
	Palette256Get(Pal1^);
	FOR i:=0 TO 32 DO BEGIN
		DataMove(Pal1^,Pal2^,768,0,0);
		Palette256Darken(Pal2^,0,255,i*2,0);
		RetraceWait;
		Palette256Set(Pal2^);
	END;
	FREEMEM(Pal1,768);
	FREEMEM(Pal2,768);
END;

PROCEDURE FadeTo(pal: POINTER);
VAR
	Pal1: POINTER;
	i: INTEGER;
BEGIN
	GETMEM(Pal1,768);
	Palette256Get(Pal1^);
	FOR i:=0 TO 63 DO BEGIN
		Palette256Transform(Pal1^,pal^);
		RetraceWait;
		Palette256Set(Pal1^);
	END;
	FREEMEM(Pal1,768);
END;

PROCEDURE DecompressRepByte(VAR InArray,OutArray; InArraySize: WORD; VAR OutArraySize: WORD);
ASSEMBLER;
ASM
	PUSH DS

	XOR DX,DX
	XOR AX,AX

	LDS SI,InArray
	LES DI,OutArray

	MOV CX,InArraySize
	JCXZ @Done

@Loop1:
	LODSB
	CMP AL,0
	JE @VsePonyatno
	CMP AL,4
	JB @MensheTreh

	INC DX
	STOSB
	JMP @DoLoop

@MensheTreh:
	SUB CX,1
	MOV BX,CX

	MOV CX,AX
	ADD DX,AX
	LODSB
	REP STOSB

	MOV CX,BX
	JMP @DoLoop

@VsePonyatno:
	LODSB
	SUB CX,2
	MOV BX,CX
	MOV CX,AX
	ADD DX,AX
	LODSB
	REP STOSB
	MOV CX,BX

@DoLoop:
	JCXZ @Done
	LOOP @Loop1

@Done:
	LES DI,OutArraySize
	MOV [ES:DI],DX
	POP DS
END;

FUNCTION MSMouseInArea(x1,y1,x2,y2: INTEGER): BOOLEAN;
ASSEMBLER;
ASM
	MOV AX,3
	INT 33H

	XOR AX,AX

	CMP CX,x1
	JLE @Done

	CMP DX,y1
	JLE @Done

	CMP CX,x2
	JGE @Done

	CMP DX,y2
	JGE @Done

	MOV AX,-1

@Done:
END;

FUNCTION MSMouseDriverExist: BOOLEAN;
ASSEMBLER;
ASM
	MOV AX,0
    INT 33H
END;

PROCEDURE MSMouseGetXY(VAR x,y: INTEGER);
ASSEMBLER;
ASM
	MOV AX,3
    INT 33H
    LES DI,x
    MOV AX,CX
    STOSW
    LES DI,y
    MOV AX,DX
    STOSW
END;

FUNCTION MSMouseButtonStatusGet: WORD;
ASSEMBLER;
ASM
	MOV AX,3
	INT 33H
	MOV AX,BX
END;

FUNCTION MSMouseButtonWasPressed(Button: WORD; VAR x,y: INTEGER): BOOLEAN;
ASSEMBLER;
ASM
	MOV AX,5
    MOV BX,Button.WORD
    INT 33H
    LES DI,x
    MOV AX,CX
    STOSW
    LES DI,y
    MOV AX,DX
    STOSW
    MOV AX,BX
END;

FUNCTION MSMouseButtonWasReleased(Button: WORD; VAR x,y: INTEGER): BOOLEAN;
ASSEMBLER;
ASM
	MOV AX,6
    MOV BX,Button.WORD
    INT 33H
    LES DI,x
    MOV AX,CX
    STOSW
    LES DI,y
    MOV AX,DX
    STOSW
    MOV AX,BX
END;

PROCEDURE MSMouseSetXY(x,y: INTEGER);
ASSEMBLER;
ASM
	MOV AX,4
	MOV CX,x
	MOV DX,y
    INT 33H
END;

PROCEDURE KeyboardFlush;
ASSEMBLER;
ASM
	MOV AX,40H
	MOV ES,AX
	MOV AX,ES:[1AH]
	MOV ES:[1CH],AX
END;

PROCEDURE DisableInterrupts;
BEGIN
	Old1B:=SetInterrupt($1B,@NullInterrupt);
	Old8:=GetInterrupt($8);
END;

PROCEDURE EnableInterrupts;
BEGIN
	SetInterrupt($1B,Old1B);
	SetInterrupt($8,Old8);
END;

PROCEDURE FHALT;
BEGIN
	EnableInterrupts;
	HALT(1);
END;

FUNCTION GetInterrupt(Int: BYTE): POINTER;
ASSEMBLER;
ASM
	MOV AL,Int
	MOV AH,35H
	INT 21H

	MOV DX,ES
	MOV AX,BX
END;

END.